//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import Alphabet;
import Distribution;
import State;
import MarkovModel;
import Common;
import Estimator;
import JointCNVDistribution;
import tango.math.Math;
import tango.io.Stdout;


//
// Create the structure of the model but dont attach meaningful distributions yet
//
MarkovModel initializeModel()
{
	// Create the model
	DiscreteAlphabet da = new DiscreteAlphabet;
	da.createSymbol("het");
	da.createSymbol("depth");
	da.createSymbol("avghit");
	MarkovModel model = new MarkovModel(da);

	// Create the states with default emission distributions
	State s_zero = new State(SNAME_ZERO, new EmptyDistribution());
	State s_norm = new State(SNAME_NORMAL, new EmptyDistribution());
	State s_cnHalf = new State(SNAME_HALF, new EmptyDistribution());
	State s_cnTwo = new State(SNAME_C2, new EmptyDistribution());
	State s_cnThree = new State(SNAME_C3, new EmptyDistribution());

	// Create the five repeat states
	State s_rptZero = new State(SNAME_RPTZERO, new EmptyDistribution());
	State s_rptNorm = new State(SNAME_RPTNORM, new EmptyDistribution());
	State s_rptHalf = new State(SNAME_RPTHALF, new EmptyDistribution());
	State s_rptTwo = new State(SNAME_RPTC2, new EmptyDistribution());
	State s_rptThree = new State(SNAME_RPTC3, new EmptyDistribution());

	// Add the states to the model
	model.addState(s_zero);
	model.addState(s_norm);
	model.addState(s_cnHalf);
	model.addState(s_cnTwo);
	model.addState(s_cnThree);

	State[] majorStates = model.getStateList();
	
	// Setup the transitions
	State terminalState = model.getTerminalState();

	// Terminal State transitions
	model.createTransition(terminalState, s_zero, 5.0);
	model.createTransition(terminalState, s_norm, 5.0);
	model.createTransition(terminalState, s_cnHalf, 1.0);
	model.createTransition(terminalState, s_cnTwo, 1.0);
	model.createTransition(terminalState, s_cnThree, 1.0);
	model.normalizeTransitions(terminalState);
	
	createInitialTransitions(model, s_zero, majorStates, s_rptZero, 1000.0, 10.0);
	createInitialTransitions(model, s_norm, majorStates, s_rptNorm, 100000.0, 10.0);
	createInitialTransitions(model, s_cnHalf, majorStates, s_rptHalf, 1000.0, 10.0);
	createInitialTransitions(model, s_cnTwo, majorStates, s_rptTwo, 1000.0, 10.0);
	createInitialTransitions(model, s_cnThree, majorStates, s_rptThree, 1000.0, 10.0);

	return model;
}

//
// Build the connections from the source state to the other top level states (non-repeat states)
// Also, set up the connection to this state's repeat
//
void createInitialTransitions(MarkovModel model, State source, 
								State[] majorStates, State repeat, 
								real selfWeight, real repeatWeight)
{
	foreach(s; majorStates)
	{
		if(source == s)
		{
			model.createTransition(source, s, selfWeight);
		}
		else
		{
			model.createTransition(source, s, 1.0);
		}
	}

	model.addState(repeat);
	model.createTransition(source, repeat, repeatWeight);
	model.createTransition(repeat, repeat, repeatWeight);
	model.createTransition(repeat, source, 1.0);
	model.normalizeTransitions(source);
	model.normalizeTransitions(repeat);
}

// update the emission distributions for the model according to the parameters
void updateModelDistributions(MarkovModel model, ModelParameters parameters)
{
	// Create the depth emission distribution
	IDistribution[char[]] depthDists;

	depthDists[SNAME_ZERO] = createExponential(parameters.stateParams[SNAME_ZERO].depth_mean);
	depthDists[SNAME_NORMAL] =	createGaussian(parameters.stateParams[SNAME_NORMAL].depth_mean, parameters.stateParams[SNAME_NORMAL].depth_sd);
	depthDists[SNAME_HALF] = createGaussian(parameters.stateParams[SNAME_HALF].depth_mean, parameters.stateParams[SNAME_HALF].depth_sd);
	depthDists[SNAME_C2] = createGaussian(parameters.stateParams[SNAME_C2].depth_mean, parameters.stateParams[SNAME_C2].depth_sd);
	depthDists[SNAME_C3] = createGaussian(parameters.stateParams[SNAME_C3].depth_mean, parameters.stateParams[SNAME_C3].depth_sd);

	// Create the repeat distributions
	foreach(rn; REPEAT_STATES)
	{
		depthDists[rn] = createExponential(parameters.stateParams[rn].depth_mean);
	}

	// Create the het emission distribution
	IDistribution[char[]] hetDists;
	hetDists[SNAME_ZERO] = createPoisson(parameters.stateParams[SNAME_ZERO].het_mean);
	hetDists[SNAME_NORMAL]  = createPoisson(parameters.stateParams[SNAME_NORMAL].het_mean);
	hetDists[SNAME_HALF]  = createPoisson(parameters.stateParams[SNAME_HALF].het_mean);
	hetDists[SNAME_C2]  = createNegBin(parameters.stateParams[SNAME_C2].het_mean, parameters.stateParams[SNAME_C2].het_sd);
	hetDists[SNAME_C3]  = createNegBin(parameters.stateParams[SNAME_C3].het_mean, parameters.stateParams[SNAME_C3].het_sd);

	// Create the repeat het distributions
	foreach(rn; REPEAT_STATES)
	{
		hetDists[rn] = createPoisson(parameters.stateParams[rn].het_mean);
	}
	
	// Create the hit emission distribution
	IDistribution[char[]] hitDists;

	hitDists[SNAME_ZERO] = createExponential(parameters.stateParams[SNAME_ZERO].avghit_mean);
	hitDists[SNAME_NORMAL] = createExponential(parameters.stateParams[SNAME_NORMAL].avghit_mean);
	hitDists[SNAME_HALF] = createExponential(parameters.stateParams[SNAME_HALF].avghit_mean);
	hitDists[SNAME_C2] = createExponential(parameters.stateParams[SNAME_C2].avghit_mean);
	hitDists[SNAME_C3] = createExponential(parameters.stateParams[SNAME_C3].avghit_mean);

	// Create the repeat het distributions
	foreach(rn; REPEAT_STATES)
	{
		hitDists[rn] = createGamma(parameters.stateParams[rn].avghit_mean, parameters.stateParams[rn].avghit_sd);
	}

	// Create the joint distributions and attach to the model
	foreach(s; model.getStateList)
	{
		if(s != model.getTerminalState())
		{
			char[] stateName = s.toString();
			IDistribution jointDist = new JointCNVDistribution(stateName, hetDists[stateName], depthDists[stateName], hitDists[stateName]);
			model.getState(stateName).setEmissionDistribution(jointDist);
		}
	}
}

//
// Create an exponential
//
ExponentialDistribution createExponential(real exp_mean)
{
	return new ExponentialDistribution(1.0 / exp_mean);
}

//
// Create a poisson exponential
//
PoissonDistribution createPoisson(real mean)
{
	return new PoissonDistribution(mean);
}


//
// Create a gamma function
//
GammaDistribution createGamma(real mean, real sd)
{
	// Calculate the shape and rate parameters for the gamma distribution
	// This is not optimal but gives a good enough fit in practice
	real var = pow(sd, 2.0);
	real scale = var / mean;
	real shape = mean / scale;
	real rate = 1 / scale;
	GammaDistribution gd = new GammaDistribution(shape, rate);
	return gd;
}

//
// Create a negative binomial distribution
//
NegBinDistribution createNegBin(real mean, real sd)
{
	// Calculate the shape and rate parameters for the gamma distribution
	// This is not optimal but gives a good enough fit in practice
	real var = pow(sd, 2.0);
	real p = 1 / (var / mean);
	real r =  mean * p / (1 - p);
	NegBinDistribution nb = new NegBinDistribution(r, p);
	return nb;
}


//
// Create a gaussian
//
GaussianDistribution createGaussian(real mean, real sd)
{
	return new GaussianDistribution(mean, sd);
}
