//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import Alphabet;
import Distribution;
import State;
import MarkovModel;
import HmmAlgorithms;
import JointCNVDistribution;
import CNVModel;
import BinData;
import Estimator;
import Common;
import StateSummary;
import tango.io.Stdout;
import Integer = tango.text.convert.Integer;
import Float = tango.text.convert.Float;
import tango.io.stream.FileStream;
import tango.text.Util;
import tango.io.stream.LineStream;
import tango.io.stream.TextFileStream;
import tango.util.ArgParser;
import tango.math.Math;

int main(char[][] args)
{
	// Set up the command line arguments and parse them
	char[] filename;
	char[] outprefix = "cnD";
	int windowSize = 1000;
	int numLearningSteps = 4;
	real smoothing = 10;
	bool bVerbose = false;
	bool bHelp = false;
	bool bNoHet = false;
	bool bHetOnly = false;

	ModelParameters modelParameters;

	ArgParser parser = new ArgParser();
	parser.bind("--", "prefix=",(char[] value){ outprefix = value; });
	parser.bind("--", "window=",(char[] value){ windowSize = Integer.parse(value); });
	parser.bind("--", "smooth=",(char[] value){ smoothing = Float.parse(value); });
	parser.bind("--", "help", { bHelp = true; });
	parser.bind("--", "nohet", { bNoHet = true; });
	parser.bind("--", "hetonly", { bHetOnly = true; });
	parser.bind("--", "verbose", { bVerbose = true; });
	parser.bind("--", "learn=", (char[] value){ numLearningSteps = Integer.parse(value); });
	parser.bindDefault((char[] value, uint i){ filename = value; });
	parser.parse(args[1..$]);	

	// The model requires at least 1 learning step
	if(numLearningSteps == 0)
	{
		numLearningSteps = 1;
	}

	if(bHelp)
	{
		printUsage();
		return 1;
	}

	if(filename == "")
	{
		Stderr("No filename passed").newline;
		printUsage();
		return 1;
	}

	if(outprefix == "")
	{
		Stderr("The outprefix cannot be the empty string").newline;
		printUsage();
		return 1;
	}

	if(windowSize <= 0)
	{
		Stderr("Window size must be greater than 0").newline;
		printUsage();
		return 1;
	}

	if(numLearningSteps <= 0)
	{
		numLearningSteps = 0;
		Stderr("Warning, no learning will be performed").newline;
	}

	// Create the output files
	TextFileOutput outFileViterbi = new TextFileOutput(outprefix ~ "_viterbi.txt");
	TextFileOutput outFileSummary = new TextFileOutput(outprefix ~ "_summary.txt");	
	TextFileOutput outFileParameters = new TextFileOutput(outprefix ~ "_parameters.txt");	
	
	// Read in the input data
	StringInputArrayMap inputDataMap = readInputData(filename);

	// Create a map of chromosomes -> symbolArray maps
	StringSymbolArrayMap chrSymArrayMap;
	foreach(char[] chr, InputArray ida; inputDataMap)
	{
		// Create the symbol array for this chromosome
		chrSymArrayMap[chr] = new SymbolArray;
		SymbolArray currArray = chrSymArrayMap[chr];

		foreach(id; ida)
		{
			real num_hets = (id.het_rate * windowSize);
			RealSymbol hrs = new RealSymbol("het", num_hets);
			RealSymbol drs = new RealSymbol("depth", id.depth);
			RealSymbol hitrs = new RealSymbol("hit", id.avg_hits);
			JointSymbol js = new JointSymbol(hrs, drs, hitrs);
			currArray.append(js);
		}
	}

	//
	// For each chromosome, learn the model and classify the states
	//
	foreach(char[] chr, SymbolArray symbolArray; chrSymArrayMap)
	{
		Stdout.formatln("Processing {} which has {} symbols", chr, symbolArray.length());
		//
		// Stage 1: Create the model and train the parameters of the entire symbol list
		//

		MarkovModel model = initializeModel();
		modelParameters = generateInitialParameters(model, symbolArray.toArray(), windowSize);
		
		// Display the initial parameters
		if(bVerbose)
		{
			Stdout("Initial Parameters: ").newline;
			modelParameters.print();
		}

		// Set the initial distributions for the model
		updateModelDistributions(model, modelParameters);

		// Clamp values to avoid underflowing the probability
		real depth_limit = 4*modelParameters.stateParams[SNAME_NORMAL].depth_mean;
		real het_limit = 4*modelParameters.stateParams[SNAME_C3].het_mean;
		foreach(s; symbolArray)
		{
			JointSymbol js = cast(JointSymbol)s;
			RealSymbol ds = js.depthSymbol;
			RealSymbol hs = js.hetSymbol;
			
			if(ds.value > depth_limit)
				ds.value = depth_limit;

			if(hs.value > het_limit)
				hs.value = het_limit;
		}

		// Learn the model parameters by iterating over Baum-Welch (transitions) and Viterbi Training (emission distributions)
		for(int i = 0; i < numLearningSteps; ++i)
		{
			if(i == 0)
			{
				// To fit the parameters, initially run the model without
				// using the het information. The het parameters will then
				// be learned and the full model will run
				JointCNVDistribution.JointCNVDistribution.disableHet();
			}

			if(bVerbose)
			{
				Stdout("\nLearning step ")(i+1).newline;
				Stdout("\nCurrent model: ").newline;
				model.printAll();
			}

			// Learn emissions from viterbi
			State[] viterbiPath = viterbi(model, symbolArray.toArray());
			updateTransitions(model, viterbiPath, smoothing);

			// Make a hash of state -> data classifications
			SymbolArray[State] stateSymbolData = classifySymbols(model, viterbiPath, symbolArray.toArray());

			// Estimate the new model parameters
			modelParameters = generateParameters(model, stateSymbolData);
			
			if(bVerbose)
			{
				Stdout("\nNew parameters: ").newline;
				modelParameters.print();
			}

			// Update the distributions
			updateModelDistributions(model, modelParameters);
			
			// Now that the parameters have been learned, enable the het channel
			if(!bNoHet)
			{
				JointCNVDistribution.JointCNVDistribution.enableHet();
			}
		}

		if(bHetOnly)
		{
			JointCNVDistribution.JointCNVDistribution.disableDepth();
			JointCNVDistribution.JointCNVDistribution.enableHet();
		}

		//
		// Stage 2: Classify all states using the model that was built
		//

		// Compute viterbi
		Stdout("Computing final viterbi path for " ~ chr).newline;	
		State[] bestPath = viterbi(model, symbolArray.toArray());
		
		// Write the viterbi decoding
		outputViterbi(outFileViterbi, chr, bestPath, inputDataMap[chr]);
		
		// Summarize the data and write it out
		SymbolArray[State] stateSymbolData = classifySymbols(model, bestPath, symbolArray.toArray());	
		outputStateSummary(outFileSummary, stateSymbolData);
	}

	outFileParameters(modelParameters.toString());

	// flush and close outfiles
	outFileViterbi.flush();
	outFileSummary.flush();
	outFileParameters.flush();
	outFileViterbi.close();
	outFileSummary.close();
	outFileParameters.close();
	return 0;
}

//
// Classify each symbol into the state that it belongs to, based on the viterbi path
//
SymbolArray[State] classifySymbols(MarkovModel model, State[] bestPath, Symbol[] symbolList)
{
	SymbolArray[State] stateSymbols;
	
	// Create the array for each state
	foreach(s; model.getStateList())
	{
		stateSymbols[s] = new SymbolArray;
	}
	
	// Add each symbol to the hash for the state it was classified to
	// the path array is 1 item longer than the symbolList since it
	// starts from the terminal state so we index at i+1
	int numSymbols = symbolList.length;
	for(int i = 0; i < numSymbols; ++i)
	{
		stateSymbols[bestPath[i+1]].append(symbolList[i]);
	}
	return stateSymbols;
}


//
//
//
void outputStateSummary(TextFileOutput outFile, SymbolArray[State] summaryData)
{
	outFile.formatln("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}", "state", "d_mean", "d_median", "num_hets_mean", "1/num_hets_mean", "ah_mean", "1 / ah_mean");

	State[] stateList = summaryData.keys.sort;
	foreach(s; stateList)
	{
		SymbolArray symbolArray = summaryData[s];

		// Columnize the values
		real[] depth_values;
		real[] het_values;
		real[] avg_hit_values;

		splitSymbolArray(symbolArray, depth_values, het_values, avg_hit_values);
	
		// Compute stats on the columns
		Statistics depth, het, avghit;
		depth = computeStats(depth_values);
		het = computeStats(het_values);
		avghit = computeStats(avg_hit_values);

		outFile.formatln("{0}\t{1:F6}\t{2:F6}\t{3:F6}\t{4:F6}\t{5:F6}\t{6:F6}", s, depth.mean,depth.median, het.mean, 1.0 / het.mean, avghit.mean, 1.0 / avghit.mean);
	}
}

//
// Output the viterbi path
//
void outputViterbi(TextFileOutput outFile, char[] label, State[] bestPath, InputArray ida)
{
	// Write out the classified data
	int numItems = ida.length;
	for(int i = 0; i < numItems; ++i)
	{
		outFile.formatln("{0}\t{1:F6}\t{2:F6}\t{3:F6}\t{4:F6}\t{5}", 
							label, ida[i].pos, ida[i].het_rate, 
							ida[i].depth, ida[i].avg_hits, bestPath[i+1]);
	}
}

//
// Read the maq window file into the data structure
//
StringInputArrayMap readInputData(char[] filename)
{
	StringInputArrayMap inputMap;
	auto input = new LineInput(new FileInput(filename));
	foreach(line; input)
	{
		if(line.length != 0)
		{
			char[][] fields = split(line, "\t");
			InputData id;
			id.chr = fields[0];
			id.pos = Float.parse(fields[1]);
			id.het_rate = Float.parse(fields[3]);
			id.depth = Float.parse(fields[4]);
			id.avg_hits = Float.parse(fields[6]);

			// Create array if necessary
			InputArray* pDA = (id.chr in inputMap);
			if(pDA == null)
				inputMap[id.chr] = new InputArray;
			inputMap[id.chr].append(id);
		}

	}
    input.close;	
	return inputMap;
}

//
// Print the usage
//
void printUsage()
{
	Stderr("Usage: cnD [OPTION] [FILE]").newline;
	Stderr("Parse FILE and call copy number variants. FILE is the output from maq cns2win.").newline;
	Stderr.newline;
	Stderr("    --prefix=STR         Use STR has the prefix for the output files").newline;
	Stderr("    --window=NUM         The window size of the input file (default: 1000)").newline;
	Stderr("    --learn=NUM          Perform Viterbi learning for NUM steps (default: 4)").newline;
	Stderr("    --smooth=NUM         Output more contiguous calls by up-weighting the self-self transition.").newline;
	Stderr("                         Higher numbers mean more contiguous calls (default: 10).").newline;
	Stderr("    --nohet              Do not use heterozygous SNP information").newline;
	Stderr("    --help               Print this message and exit").newline;
	Stderr("    --verbose            Print status and model parameters during processing").newline;
}
