//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
module Distribution;
import Alphabet;
import tango.io.Stdout;
import Float = tango.text.convert.Float;
import tango.math.random.Kiss;
import tango.math.Math;
import tango.math.Probability;
import tango.math.GammaFunction;
import tango.util.collection.ArraySeq;

// Implementaton of probability distributions over discrete and continuous variables
// Interface inspired by the BioJava (biojava.org) project
interface IDistribution
{
	// Get the weight of the given symbol (from the PDF or PMF)
	real getWeight(Symbol s);

	// Get the probability (from the CDF) that X is less than x
	real getProbLessThan(Symbol s);

	void printWeight(Symbol s);

	// Sample a symbol from the distribution
	Symbol sample();

	// Does the distribution contain this symbol
	bool contains(Symbol s);

	char[] toString();
}

//
// A probability distribution over a collection of discrete symbols
//
class DiscreteDistribution : IDistribution
{
	this() {}
	this(DiscreteAlphabet da)
	{
		foreach(n,s; da)
		{
			addSymbol(s);
		}
	}

	//
	// Add a symbol to the distribution
	//
	void addSymbol(Symbol s)
	{
		m_weights[s] = 0.0f;
	}

	//
	// Check if the distribution contains this symbol
	//
	bool contains(Symbol s)
	{
		real* p = (s in m_weights);
		return p != null;
	}
	
	//
	// Set the weight of a particular symbol
	//
	void setWeight(Symbol s, real value)
	{
		real* p = (s in m_weights);
		if(p == null)
		{
			throw new Exception("Distributuin::setWeight - Symbol not found in distribution: " ~ s.toString());
		}
		*p = value;
	
		// Update the cumulative distribution
		generateCumulative();
	}

	//
	//
	//
	void printWeight(Symbol s)
	{	
	}

	
	//
	// Get the weight in the pmf of the passed in symbol
	//
	real getWeight(Symbol s)
	{
		real* p = (s in m_weights);
		if(p == null)
		{
			throw new Exception("Distribution::getWeight - Symbol not found in distribution: " ~ s.toString());
		}
		return *p; 
	}

	//
	// From the cumulative distribution, return the probability that X is less than or equal to x
	//
	real getProbLessThan(Symbol s)
	{
		// This is lame
		foreach(cn; m_cumulativeDist)
		{
			if(cn.sym == s)
			{
				return cn.endpoint;
			}
		}

		return 0.0f; 
	}

	//
	// Normalize the distribution weights
	//
	void normalize()
	{
		real sum = 0.0f;
		foreach(s,w; m_weights)
		{
			sum += w;
		}

		foreach(s,w; m_weights)
		{
			m_weights[s] = w / sum;
		}
	}
	
	//
	// Generate the cumulative distribution over the symbols
	//
	void generateCumulative()
	{
		m_cumulativeDist.length = m_weights.length;
		int index = 0;
		real sum = 0.0;
		foreach(s,w; m_weights)
		{
			sum += w;
			m_cumulativeDist[index].endpoint = sum;
			m_cumulativeDist[index].sym = s;
			++index;
		}
	}

	//
	// Sample a symbol from the distribution
	//
	Symbol sample()
	{
		// Generate a random number
		auto r = Kiss.shared.toReal();

		// Find the place in the cdf (this could be faster but is good enough for small alphabets
		foreach(cn; m_cumulativeDist)
		{
			if(r < cn.endpoint)
			{
				return cn.sym;
			}
		}
		throw new Exception("Attempted to sample out of bounds (" ~ Float.toString(r) ~ ")");
	}
	
	//
	// 
	//
	char[] toString()
	{
		char[] outStr;
		foreach(s,w; m_weights)
		{
			outStr ~= s.toString() ~ " : " ~ Float.toString(1/w)  ~ "\t";
		}
		return outStr;
	}

	//
	//
	//
	void printCumulative()
	{
		foreach(cn; m_cumulativeDist)
		{
			Stdout(cn.sym)(":")(cn.endpoint).newline;
		}
	}


	//
	// Members
	//
	struct CumulativeDistNode
	{
		real endpoint;
		Symbol sym;
	}

	real[Symbol] m_weights;
	CumulativeDistNode[] m_cumulativeDist;
}

//
// Gaussian Distribution
//
class GaussianDistribution : IDistribution
{
	this(real mean, real sd) { m_mean = mean; m_sd = sd; }

	// Evaluate the pdf at the passed in value (of the realsymbol)
	real getWeight(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to GaussianDistribution(pdf): " ~ s.classinfo.name);
		}
		
		RealSymbol rs = cast(RealSymbol)s;
		
		// Calculate the probability of this value
		real a = 1.0 / (m_sd * sqrt(2*PI));
		real l = -(pow((rs.value - m_mean), 2) / (2*(pow(m_sd, 2.0))));
		real p = a * exp(l);
		
		return p;
	}

	// Evaluate the CDF at the value
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to GaussianDistribution(cdf): " ~ s.classinfo.name);
		}
		
		// Check if the distribution is valid
		if(m_sd > 0.0f)
		{
			RealSymbol rs = cast(RealSymbol)s;
			return normalDistribution((rs.value - m_mean) / m_sd); 
		}
		else
		{
			return 0;
		}
	}

	// Sample a value from the distribution
	RealSymbol sample()
	{
		auto r = Kiss.shared.toReal();
		real p = normalDistributionInv(r) * m_sd + m_mean;
		return new RealSymbol("", p);
	}

	// It contains all symbols, return true
	bool contains(Symbol s)
	{
		return true;
	}

	//
	//
	//
	void printWeight(Symbol s)
	{	
	}

	//
	//
	//
	char[] toString()
	{
		return "N(" ~ Float.toString(m_mean) ~ "," ~ Float.toString(m_sd*m_sd) ~ ")";
	}

	real m_mean;
	real m_sd;
}

//
// A mixture of weighted distribution
//
class MixedDistribution : IDistribution
{
	this() { m_distributions = new DataArray; }

	// add a gaussian with the specified weight
	void addDistribution(real w, IDistribution dist)
	{
		MixedData md;
		md.weight = w;
		md.dist = dist;
		m_distributions.append(md);
	}

	// Sum the weighted p values
	real getWeight(Symbol s)
	{
		real sum = 0.0f;
		foreach(md; m_distributions)
		{
			sum += md.weight * md.dist.getWeight(s);
		}
		
		return sum;
	}

	// Sum the weighted probabilities
	real getProbLessThan(Symbol s)
	{
		real sum = 0.0f;
		foreach(md; m_distributions)
		{
			sum += md.weight * md.dist.getProbLessThan(s);
		}
		return sum;
	}

	// Sampling not implemented
	RealSymbol sample()
	{
		throw new Exception("Sampling not implemented for mixed distribution");
		return new RealSymbol("", 0.0f);
	}

	// Contains all symbols
	bool contains(Symbol s)
	{
		return true;
	}

	//
	//
	//
	char[] toString()
	{
		return "MixedDistribution";
	}
	
	
	// A data point structure
	struct MixedData
	{
		real weight;
		IDistribution dist;
	}

	//
	//
	//
	void printWeight(Symbol s)
	{	
	}
	

	alias ArraySeq!(MixedData) DataArray;
	DataArray m_distributions;
}


//
// A distribution taking continuous values that are quantized to discrete values
//
class QuantizedDistribution : IDistribution
{
	this(real step, IDistribution dist)
	{
		m_stepSize = step;
		m_dist = dist;
	}

	// Get the weight (probability in this case) of the value of the symbol
	real getWeight(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to QuantiziedDistribution: " ~ s.classinfo.name);
		}
		
		RealSymbol rs = cast(RealSymbol)s;
		real value = rs.value;

		// Calculate the interval endpoints
		real start = floor(value / m_stepSize) * m_stepSize;
		real end = start + m_stepSize;
		
		real p = m_dist.getProbLessThan(new RealSymbol("", end)) - m_dist.getProbLessThan(new RealSymbol("", start));
		if(p == real.nan)
		{
			throw new Exception("Distribution return NaN probability");
		}
		return p;
	}

	// Get the probabily <= x
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to QuantizedDistribution: " ~ s.classinfo.name);
		}
		
		RealSymbol rs = cast(RealSymbol)s;
		real start = floor(rs.value / m_stepSize) * m_stepSize;
		real end = start + m_stepSize;
		return m_dist.getProbLessThan(new RealSymbol("", end));
	}

	// Sampling not allowed
	RealSymbol sample()
	{
		throw new Exception("Sampling not implemented for gaussian mixture");
		return new RealSymbol("", 0.0f);
	}

	// Contains all symbols by default
	bool contains(Symbol s)
	{
		return true;
	}

	//
	//
	//
	void printWeight(Symbol s)
	{	
	}

	//
	//
	//
	char[] toString()
	{
		return "QuantizedDistribution";
	}

	
	real m_stepSize; // the distance between quants
	IDistribution m_dist;
}

//
// Poisson distribution
//
class PoissonDistribution : IDistribution
{
	this(real lambda) { m_lambda = lambda; }

	// Get the weight (probability) of the value passed in
	real getWeight(Symbol s)
	{
		int k;
		if(s.classinfo == IntSymbol.classinfo)
		{
			IntSymbol rs = cast(IntSymbol)s;
			k = rs.value;
		}
		else if(s.classinfo == RealSymbol.classinfo)
		{
			RealSymbol rs = cast(RealSymbol)s;
			k = cast(int)round(rs.value);
		}
		else
		{
			throw new Exception("Invalid symbol type to PoissonDistribution: " ~ s.classinfo.name);
		}
		
		if(k == 0)
		{
			return poissonDistribution(k, m_lambda);
		}
		else
		{
			return poissonDistribution(k, m_lambda) - poissonDistribution(k-1, m_lambda) ;
		}
	}

	// Get the probability that the value is <= x
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != IntSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to PoissonDistribution: " ~ s.classinfo.name);
		}
		
		IntSymbol rs = cast(IntSymbol)s;
		uint k = rs.value;		
		return poissonDistribution(k, m_lambda);
	}
	
	// Sample from the possion dist
	IntSymbol sample()
	{
		auto r = Kiss.shared.toReal();
		
		int k = 0;
		while(r > poissonDistribution(k, m_lambda))
		{
			++k;
		}
		
		return new IntSymbol("", k);
	}

	//
	bool contains(Symbol s)
	{
		return true;
	}

	//
	void printWeight(Symbol s)
	{	
	}

	//
	char[] toString()
	{
		return "P(" ~ Float.toString(m_lambda) ~ ")";
	}

	real m_lambda;
}

//
// Exponential distribution
//
class ExponentialDistribution : IDistribution
{
	this(real lambda) 
	{
		if(m_lambda == real.nan || m_lambda == real.infinity)
		{
			Stderr("Warning creating degenerate exponential distribution with parameter ")(lambda).newline;
		}
		m_lambda = lambda; 
	}

	// Evaluate the pdf at the position
	real getWeight(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to ExponentialDistribution: " ~ s.classinfo.name);
		}

		
		if(m_lambda != real.nan && m_lambda != real.infinity)
		{
			RealSymbol rs = cast(RealSymbol)s;
			real x = rs.value;
			return m_lambda * exp(-(m_lambda*x));
		}
		else
		{
			return 0.0f;
		}
		
	}

	// Evaluate the cdf at the value
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to ExponentialDistribution: " ~ s.classinfo.name);
		}
		
		RealSymbol rs = cast(RealSymbol)s;	
		return cdf(rs.value);
	}

	// Sample a symbol
	RealSymbol sample()
	{
		auto r = Kiss.shared.toReal();
		
		real x = (-log(1 - r)) / m_lambda;
		return new RealSymbol("", x);
	}

	// CDF Implementation
	real cdf(real x)
	{
		if(m_lambda != real.nan && m_lambda != real.infinity)
		{
			return (1.0 - exp(-(m_lambda*x)));
		}
		else
		{
			return 0.0f;
		}
	}

	// Contains all symbols
	bool contains(Symbol s)
	{
		return true;
	}

	//
	void printWeight(Symbol s)
	{	
	}
	
	//
	char[] toString()
	{
		return "E(" ~ Float.toString(m_lambda) ~ ")";
	}
	

	real m_lambda;
}

//
// Gamma distribution
//
class GammaDistribution : IDistribution
{
	this(real shape, real rate) 
	{
		if(shape == real.nan || shape == real.infinity || rate == real.nan || rate == real.infinity)
		{
			throw new Exception("Invalid parameters to GammaDistribution: " ~ Float.toString(shape) ~ ", " ~ Float.toString(rate));
		}
		m_rate = rate; 
		m_shape = shape;
	}

	// Evaluate the pdf at the position
	real getWeight(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to GammaDistribution: " ~ s.classinfo.name);
		}

		RealSymbol rs = cast(RealSymbol)s;
		real x = rs.value;
		real p = pow(m_rate, m_shape) * pow(x, m_shape - 1) * exp(-m_rate * x) / gamma(m_shape);
		return p;
	}

	// Evaluate the cdf at the value
	real getProbLessThan(Symbol s)
	{
		if(s.classinfo != RealSymbol.classinfo)
		{
			throw new Exception("Invalid symbol type to ExponentialDistribution: " ~ s.classinfo.name);
		}
		
		RealSymbol rs = cast(RealSymbol)s;	
		return cdf(rs.value);
	}

	// Sample a symbol
	RealSymbol sample()
	{
		throw new Exception("Sampled not implemented for gamma distribution");
		auto r = Kiss.shared.toReal();
		return new RealSymbol("", r);
	}

	// CDF Implementation
	real cdf(real x)
	{
		return gammaDistribution(m_rate, m_shape, x);
	}

	// Contains all symbols
	bool contains(Symbol s)
	{
		return true;
	}

	//
	void printWeight(Symbol s)
	{	
	}

	//
	char[] toString()
	{
		return "G(" ~ Float.toString(m_shape) ~ "," ~ Float.toString(m_rate) ~ ")";
	}
	

	real m_shape;
	real m_rate;
}

//
// Negative Binomial distribution
//
class NegBinDistribution : IDistribution
{
	this(real r, real p) 
	{
		if(r == real.nan || r == real.infinity || p == real.nan || p == real.infinity)
		{
			throw new Exception("Invalid parameters to NegativeBinomial: " ~ Float.toString(r) ~ ", " ~ Float.toString(p));
		}
		m_r = r; 
		m_p = p;
	}

	// Get the weight (probability) of the value passed in
	real getWeight(Symbol s)
	{
		int k;
		if(s.classinfo == IntSymbol.classinfo)
		{
			IntSymbol rs = cast(IntSymbol)s;
			k = rs.value;
		}
		else if(s.classinfo == RealSymbol.classinfo)
		{
			RealSymbol rs = cast(RealSymbol)s;
			k = cast(int)round(rs.value);
		}
		else
		{
			throw new Exception("Invalid symbol type to NegBinDistribution: " ~ s.classinfo.name);
		}
		
		if(k == 0)
		{
			return betaIncomplete( m_r, k + 1, m_p );
		}
		else
		{
			return betaIncomplete( m_r, k + 1, m_p ) - betaIncomplete( m_r, k, m_p );
		}
	}

	// Get the probability that the value is <= x
	real getProbLessThan(Symbol s)
	{
		int k;
		if(s.classinfo == IntSymbol.classinfo)
		{
			IntSymbol rs = cast(IntSymbol)s;
			k = rs.value;
		}
		else if(s.classinfo == RealSymbol.classinfo)
		{
			RealSymbol rs = cast(RealSymbol)s;
			k = cast(int)round(rs.value);
		}
		else
		{
			throw new Exception("Invalid symbol type to NegBinDistribution: " ~ s.classinfo.name);
		}
		return betaIncomplete( m_r, k + 1, m_p );
	}
	
	// Sample from the possion dist
	IntSymbol sample()
	{
		throw new Exception("Sampled not implemented for negbin distribution");
		return new IntSymbol("", 0);
	}

	bool contains(Symbol s)
	{
		return true;
	}
	
	//
	void printWeight(Symbol s)
	{	
	}

	//
	char[] toString()
	{
		return "NB(" ~ Float.toString(m_r) ~ "," ~ Float.toString(m_p) ~ ")";
	}
	

	real m_r;
	real m_p;
}


// Factorial for poisson
int factorial(int k)
{
	if(k <= 0)
	{
		return 1;
	}
	int r = 1;
	for(int i = k; k > 1; k--)
	{
		r *= k;
	}
	return r;
}

//
// Used for the termainl state, does not emit symbols
//
class EmptyDistribution : IDistribution
{
	real getWeight(Symbol s) { return 0.0f; }
	real getProbLessThan(Symbol s) { return 0.0f; }
	Symbol sample()
	{
		throw new Exception("Sampled from an empty distribution");
		return new Symbol("");
	}

	//
	void printWeight(Symbol s)
	{	
	}
	
	char[] toString()
	{
		return "EmptyDistribution";
	}

	bool contains(Symbol s) { return false; }
}
