//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
import Alphabet;
import Distribution;
import State;
import MarkovModel;
import tango.io.Stdout;
import tango.util.collection.ArraySeq;
import tango.util.collection.HashMap;
import tango.math.Math;
import Float = tango.text.convert.Float;
import tango.math.random.Kiss;
import tango.io.stream.TextFileStream;

// Debug flag
int DEBUG_VITERBI = 0;

//
// Aliases
//

alias HashMap!(Symbol, real) SymbolRealMap;
alias HashMap!(State, State) SymbolSymbolMap;
alias ArraySeq!(Sample) SampleArray;
alias ArraySeq!(State) StatePath;
alias SymbolRealMap[Symbol] LearnMatrix;

//
// Structs
//

// A sample from a markov model
struct Sample
{
	char[] stateStr;
	Symbol symbol;
}

struct DPResult
{
	real score;
	SymbolRealMap[] matrix;
}


//
// Generate samples from a given Markov model
//
class MMSampler
{
	this(MarkovModel m)
	{
		m_model = m;
		m_currState = m_model.getTerminalState();
	}

	SampleArray generate(int numSamples)
	{
		// Uncomment this line to seed the RNG with a fixed value
		// Useful for debugging when you want a fixed set of samples
		//Stdout("WARNING RNG SEED FIXED").newline;
		//Kiss.shared.seed(3);
		
		SampleArray generated = new SampleArray();
		for(int i = 0; i < numSamples; ++i)
		{
			// Get a state to transition to
			// loop until a new state is selected that isnt the terminal state (we want a specific number of samples)
			do
			{
				m_currState = cast(State)m_model.getTransitionWeights(m_currState).sample();
			} 
			while (m_currState == m_model.getTerminalState())			
			
			// Emit a symbol
			Symbol emission = m_currState.getEmissionDistribution().sample();
			Sample n;
			
			n.stateStr = m_currState.toString();
			n.symbol = emission;
			generated.append(n);
		}
		return generated;
	}
	
	State m_currState;
	MarkovModel m_model;
}

// Compute the viterbi path over the model given observations
State[] viterbi(MarkovModel model, Symbol[] observations)
{
	SymbolRealMap[] mpp; // Most probable path ending at State
	SymbolSymbolMap[] bestPrevState; // The best state transitioning to l from k at i-1 
	
	State terminal = model.getTerminalState();

	mpp.length = observations.length + 1;
	bestPrevState.length = observations.length + 1;

	// Initialize the StateMaps
	for(int i = 0; i < mpp.length; ++i)
	{
		mpp[i] = new SymbolRealMap();
		bestPrevState[i] = new SymbolSymbolMap();
	}
	
	// Save the list of all states in the model
	State[] stateList = model.getStateList();
	
	//
	// Initialize the most probable paths
	//
	foreach(s; stateList)
	{
		// the initial states are impossible
		mpp[0][s] = -real.infinity;
		//bestPrevState[0][s] = model.getTerminalState();
	}
	
	mpp[0][terminal] = vbLog(1.0f); //Must start in initial state

	//
	// Calculate the most probable paths going forward
	//
	for(int i = 0; i < observations.length; ++i)
	{
		int idx = i + 1;
		SymbolRealMap prevV = mpp[idx - 1];
		if(DEBUG_VITERBI > 1) Stdout("Obs idx ")(idx).newline;
		// Update the mpp for each state
		foreach(l; stateList)
		{
			//if(DEBUG_VITERBI) Stdout("Updating Vk for ")(l).newline;
			// Find the max vk * akl for this state
			real maxlp = -real.infinity;
			State bestState = null;
			foreach(k; stateList)
			{
				IDistribution k_transitions = model.getTransitionWeights(k);

				if(k_transitions.contains(l)) // does the state k have a transition to l
				{
					// prevVK is in log space already, add the log of the transition prob from k to l
					real currlp = prevV[k] + vbLog(k_transitions.getWeight(l));
					//if(DEBUG_VITERBI > 1) Stdout(k)(" Curr: ")(currlp)(" Max: ")(maxlp).newline;
					if(currlp > maxlp)
					{
						maxlp = currlp;
						bestState = k;
					}
				}
			}
			
			if(maxlp > -real.infinity)
			{
				bestPrevState[idx][l] = bestState;
				IDistribution emissionDist = l.getEmissionDistribution();
				
				if(emissionDist.contains(observations[i]))
				{
					mpp[idx][l] = vbLog(emissionDist.getWeight(observations[i])) + maxlp;
				}
				else
				{
					mpp[idx][l] = -real.infinity;
				}
			}
			else
			{
				mpp[idx][l] = -real.infinity;
			}
		}
	}

	// Terminate and calculate the probabilty of the full path
	real lpFinal = -real.infinity;
	State[] bestPath;
	bestPath.length = observations.length + 1;
	
	// Calculate Vk for all the states
	foreach(k; stateList)
	{
		IDistribution k_transitions = model.getTransitionWeights(k);

		if(k_transitions.contains(terminal)) // does the state k have a transition to l
		{
			real curr = mpp[observations.length][k] + vbLog(model.getTransitionWeights(k).getWeight(terminal));	
			
			if(DEBUG_VITERBI > 1) Stdout("State ")(k)(" val ")(curr)(" lpf ")(lpFinal).newline;

			if(curr > lpFinal)
			{
				lpFinal = curr;
				bestPath[observations.length] = k;
			}
		}
	}

	if(DEBUG_VITERBI > 1) Stdout("Last state set to: ")(bestPath[observations.length]).newline;

	// trace backwards generating the best state path
	for(int i = observations.length; i > 0; --i)
	{
		int index = i - 1;
		bestPath[index] = bestPrevState[i][bestPath[i]];
		if(DEBUG_VITERBI >= 1) Stdout("index: ")(index)(" Prev: ")(bestPath[index])(" Curr: ")(bestPrevState[i][bestPath[i]]).newline;
	}
	//bestPath[0] = model.getTerminalState();
	return bestPath;
}

DPResult forward(MarkovModel model, Symbol[] observations)
{
	// Sum of probabilities ending at State
	SymbolRealMap[] sum_f; // Most probable path ending at State
	
	// For convienance, save a reference to the terminal state
	State terminal = model.getTerminalState();
	// Save the list of all states in the model
	State[] stateList = model.getStateList();

	sum_f.length = observations.length + 1;

	// Initialize the StateMaps
	for(int i = 0; i < sum_f.length; ++i)
	{
		sum_f[i] = new SymbolRealMap();
	}
	
	//
	// Initialize the sum
	//
	foreach(s; stateList)
	{
		// the initial states are impossible
		sum_f[0][s] = -real.infinity;
		//bestPrevState[0][s] = model.getTerminalState();
	}
	
	sum_f[0][terminal] = vbLog(1.0f); //Must start in initial state

	for(int i = 0; i < observations.length; ++i)
	{
		int idx = i + 1;
	//	Stdout.format("Index {}", idx).newline;
		
		SymbolRealMap prevF = sum_f[idx - 1];
		if(DEBUG_VITERBI > 1) Stdout("Obs idx ")(idx)(" Observation ")(observations[i]).newline;
		
		// Update sum_f for each state
		foreach(l; stateList)
		{

			real max_fk = pickMaxInState(prevF, stateList);
			
			if(DEBUG_VITERBI > 1) Stdout("State: ")(l)(" ");
			real new_sum_l = 0.0f;
			foreach(k; stateList)
			{
				if(DEBUG_VITERBI > 1) Stdout(k)("->");
				IDistribution k_transitions = model.getTransitionWeights(k);
				
				// Do not add to the sum if there is not a transition from k to l
				// or the probability of being in the previous state is zero
				if(k_transitions.contains(l) && prevF[k] != -real.infinity) // does the state k have a transition to l
				{
					real a_kl = vbLog(k_transitions.getWeight(l));
					//new_sum_l += exp(prevF[k])*k_transitions.getWeight(l);
					new_sum_l += exp(a_kl + (prevF[k] - max_fk)); 
					//new_sum_l += exp(prevF[k] + vbLog(k_transitions.getWeight(l)));
					//Stdout.format("Calculating score ({0:F16},{1:F16}), t: {2:F16} constant: {3:F16} prev: {4:F16} sum {5:F16}\n", l, k, a_kl, max_fk, prevF[k], new_sum_l);
					
				}
				if(DEBUG_VITERBI > 1) Stdout(new_sum_l)(" ");
			}

			// calculate the final value for this (idx,l)
			real sym_prob = l.getEmissionDistribution().getWeight(observations[i]);
			real leli = vbLog(sym_prob);
			sum_f[idx][l] = leli + vbLog(new_sum_l) + max_fk;
			//Stdout.format("Setting col: {0:F16} sym_prob: {1:F16} log prob: {2:F16} log score {3:F16}\n", sum_f[idx][l], sym_prob, leli, vbLog(new_sum_l));
			//sum_f[idx][l] = vbLog(l.getEmissionDistribution().getWeight(observations[i])) + vbLog(new_sum_l);
			if(DEBUG_VITERBI > 1) Stdout("=")(sum_f[idx][l]).newline;
		}

		// Check for a state wipe
		bool allInf = true;
		foreach(l; stateList)
		{
			if(sum_f[idx][l] != -real.infinity)
			{
				allInf = false;
				break;
			}
		}

		if(allInf)
		{
			printStateWipeMessage(model, observations[i], i);
			throw new Exception("The data cannot be explained by the model, aborting");
		}
		
	}

	// Termination
	real p_x = 0.0f;
	real new_sum = 0.0f;

	SymbolRealMap prevF = sum_f[observations.length];
	real maxFp = pickMaxInState(prevF, stateList);	
	
	foreach(k; stateList)
	{
		if(prevF[k] != -real.infinity)
		{
			if(model.hasTransition(k, terminal))
			{
				new_sum += exp(vbLog(model.getTransitionWeights(k).getWeight(terminal)) + (prevF[k] - maxFp));
			}
		}
	}

	p_x = vbLog(new_sum) + maxFp;

	DPResult result;
	result.matrix = sum_f;
	result.score = p_x;
	return result;
}

DPResult backward(MarkovModel model, Symbol[] observations)
{
	SymbolRealMap[] sum_b; 
	
	// For convienance, save a reference to the terminal state
	State terminal = model.getTerminalState();
	// Save the list of all states in the model
	State[] stateList = model.getStateList();

	sum_b.length = observations.length + 1;

	// Initialize the StateMaps
	for(int i = 0; i < sum_b.length; ++i)
	{
		sum_b[i] = new SymbolRealMap();
	}
	
	//
	// Initialize the sum
	//
	int lastIdx = sum_b.length - 1;
	foreach(s; stateList)
	{
		// the initial states are impossible
		if(s != terminal && model.hasTransition(s, terminal))
		{
			sum_b[lastIdx][s] = vbLog(model.getTransitionWeights(s).getWeight(terminal));
		}
		else
		{
			sum_b[lastIdx][s] = -real.infinity;
		}
	}
	
	//
	// Recursion
	//
	for(int i = observations.length - 1; i >= 0; --i)
	{
		int idx = i;
	//	Stdout.format("Index {}", idx).newline;
		
		SymbolRealMap prevB = sum_b[idx + 1];
		if(DEBUG_VITERBI > 1) Stdout("Obs idx ")(idx).newline;

		// Update sum_b for each state
		foreach(k; stateList)
		{
			IDistribution k_transitions = model.getTransitionWeights(k);
			real max_bl = pickMaxInState(prevB, stateList);
			
			if(DEBUG_VITERBI > 1) Stdout("State: ")(k)(" ");
			real new_sum_l = 0.0f;
			
			foreach(l; stateList)
			{
				IDistribution l_emissions = l.getEmissionDistribution();
				
				if(DEBUG_VITERBI > 1) Stdout(k)("->");
				
				// Do not add to the sum if there is not a transition from k to l
				// or the probability of being in the previous state is zero
				if(k_transitions.contains(l) && prevB[l] != -real.infinity) // does the state k have a transition to l
				{
					real a_kl = vbLog(k_transitions.getWeight(l));
					real e_lo = vbLog(l_emissions.getWeight(observations[idx]));
					//new_sum_l += exp(prevF[k])*k_transitions.getWeight(l);
					new_sum_l += exp(a_kl + e_lo + (prevB[l] - max_bl)); 
					//new_sum_l += exp(prevF[k] + vbLog(k_transitions.getWeight(l)));
					//Stdout.format("Calculating score ({0:F16},{1:F16}), t: {2:F16} constant: {3:F16} prev: {4:F16} sum {5:F16}\n", l, k, a_kl, max_fk, prevF[k], new_sum_l);
					
				}
				if(DEBUG_VITERBI > 1) Stdout(new_sum_l)(" ");
			}

			// calculate the final value for this (idx,l)
			sum_b[idx][k] = vbLog(new_sum_l) + max_bl;
			//Stdout.format("Setting col: {0:F16} weight: {1:F16} log score {2:F16}\n", sum_f[idx][l], leli, vbLog(new_sum_l));
			//sum_f[idx][l] = vbLog(l.getEmissionDistribution().getWeight(observations[i])) + vbLog(new_sum_l);
			if(DEBUG_VITERBI > 1) Stdout("=")(sum_b[idx][k]).newline;
		}
	}
	
	//
	// Termination
	//
	real p_x = 0.0f;
	real new_sum = 0.0f;

	SymbolRealMap prevB = sum_b[1];
	real maxBp = pickMaxInState(prevB, stateList);	
	
	foreach(l; stateList)
	{
		if(l != terminal && prevB[l] != -real.infinity && model.hasTransition(terminal, l))
		{
			real a_tl = vbLog(model.getTransitionWeights(terminal).getWeight(l));
			real e_l = vbLog(l.getEmissionDistribution().getWeight(observations[0]));
			new_sum += exp(a_tl + e_l + (prevB[l] - maxBp));
		}
	}

	p_x = vbLog(new_sum) + maxBp;
	DPResult result;
	result.matrix = sum_b;
	result.score = p_x;
	return result;
	
}

//
// Generate the posterior state probabilities
//

void writePosteriorState(TextFileOutput outFile, DPResult forwardResult, DPResult backwardResult)
{
	SymbolRealMap[] f_srm = forwardResult.matrix;
	SymbolRealMap[] b_srm = backwardResult.matrix;
	real score = forwardResult.score; // == backwardResult.score
	int numStates = f_srm.length; // == b_srm.length by definition

	// Print a header
	outFile.format("Index\t");
	foreach(state, sc; f_srm[0])
	{
		outFile.format("{}\t", state);
	}
	outFile.newline;
	
	for(int i = 0; i < numStates; ++i)
	{
		outFile.format("{}\t", i);
		
		foreach(state, sc; f_srm[i])
		{
			real post_prob = exp(f_srm[i][state] + b_srm[i][state] - score);
			outFile.format("{0:F6}\t", post_prob);
		}

		outFile.newline;
	}
}

//
// Baum-Welch unsupervised training algorithm
//
double baumWelch(MarkovModel model, Symbol[] observations, int numIters)
{
	// Assume model is primed with initial parameters

	// Cache the state and emission list
	State[] stateList = model.getStateList();
	Symbol[] emissionList = model.getEmissionList();

	// Create the transition count matrices
	LearnMatrix A_kl;
	real score;
	foreach(k; stateList)
	{
		A_kl[k] = new SymbolRealMap;
	}

	//printLM(A_kl);
	//printLM(B_ke);
	
	for(int iter = 0; iter < numIters; ++iter)
	{
		// Initialize A and B
		foreach(k; stateList)
		{
			foreach(l; stateList)
			{
				if(model.hasTransition(k, l))
				{
					A_kl[k][l] = 1.0;
				}
			}
		}
		
		// Calculate forward and backward matrices
		DPResult forwardDPResult = forward(model, observations);
		DPResult backwardDPResult = backward(model, observations);
		SymbolRealMap[] f_m = forwardDPResult.matrix;
		SymbolRealMap[] b_m = backwardDPResult.matrix;
		score = forwardDPResult.score;
		//Stdout.format("Curr score: {0:F10}\n", score);

		// Update A_kl
		foreach(k; stateList)
		{
			foreach(l; stateList)
			{
				if(model.hasTransition(k,l))
				{
					real akl = vbLog(model.getTransitionWeights(k).getWeight(l));
					IDistribution emis_l = l.getEmissionDistribution();
					real tempscore = 0.0;
					
					for(int idx = 0; idx < observations.length; ++idx)
					{
						real e_li = emis_l.getWeight(observations[idx]);
						if(e_li != -real.infinity && f_m[idx][k] != -real.infinity && b_m[idx][l] != -real.infinity)
						{
							tempscore += exp(f_m[idx][k] + vbLog(e_li) + akl + b_m[idx + 1][l] - score);
						}
					}

					A_kl[k][l] = A_kl[k][l] + tempscore;
				}
			}
		}

		//printLM(A_kl);

		//Calculate the new transition
		foreach(k; stateList)
		{
			DiscreteDistribution k_transitions = model.getTransitionWeights(k);
			
			real sum = 0.0;
			foreach(l; stateList)
			{
				if(model.hasTransition(k, l))
				{
					sum += A_kl[k][l];
				}
			}

			foreach(l; stateList)
			{
				if(model.hasTransition(k,l))
				{
					real num_t = A_kl[k][l] / sum;
					k_transitions.setWeight(l, num_t);
					//Stdout.format("Trans: ({},{}) = {2:F6}\n", k, l, num_t);
				}
			}
		}
	}
	return score;
	//model.printAll();

}

//
// Helpers
//
real pickMaxInState(SymbolRealMap srMap, State[] stateList)
{
	real max_fk = -real.infinity;
	foreach(k; stateList)
	{
		if(srMap[k] > max_fk)
		{
			max_fk = srMap[k];
		}
	}
	return max_fk;
}

// Wrapper for the log function
real vbLog(real p)
{
	return log(p);
}

// Print a transition learning matrix
void printLM(LearnMatrix tlm)
{
	foreach(k, hm; tlm)
	{
		foreach(l,v; hm)
		{
			Stdout(k)(",")(l)(" = ")(v).newline;
		}
	}
}

// Print a dp result matrix
void printDPResult(DPResult result)
{
	SymbolRealMap[] srm = result.matrix;
	foreach(i, rm; srm)
	{
		Stdout(i)(": ");
		
		foreach(s,p; rm)
		{
			Stdout.format("{}->{1:F16}\t", s, p);
		}

		Stdout.newline;
	}		
	Stdout.format("P(x) = {}\n", result.score);
}

//
// Output debug info in the case that the model becomes impossible
//
void printStateWipeMessage(MarkovModel model, Symbol observation, int index)		
{
	Stdout("Fatal error, all state probabilities are -infinity, dumping into\nData idx: ")(index).newline;
	foreach(l; model.getStateList())
	{
		IDistribution l_emissions = l.getEmissionDistribution();
		Stdout(l.toString())(" ");
		l_emissions.printWeight(observation);
	}

}
