//-----------------------------------------------
// Copyright 2009 Wellcome Trust Sanger Institute
// Written by Jared Simpson (js18@sanger.ac.uk)
// Released under the GPL license
//-----------------------------------------------
module MarkovModel;
import Alphabet;
import Distribution;
import State;
import tango.io.Stdout;
import tango.util.container.HashSet;
import tango.io.stream.FileStream;
import tango.text.Util;
import tango.io.stream.LineStream;
import tango.io.stream.TextFileStream;



// Implementaton of a simple markov model
// Interface inspired by the BioJava (biojava.org) project

class MarkovModel
{

	//
	//
	//
	this(IAlphabet symbolAlpha)
	{
		m_symbolAlphabet = symbolAlpha;
		
		m_terminal = new State("TerminalState", new EmptyDistribution);
		addState(m_terminal);
	}

	//
	//
	//
	IAlphabet getSymbolAlphabet() { return m_symbolAlphabet; }

	//
	//
	//
	void addState(State s)
	{
		// Ensure the state doesnt already exist
		State* pState = (s.toString() in m_stateLUT);
		if(pState != null)
		{
			throw new Exception("State " ~ s.toString ~ " already exists");
		}

		// Add the state to the lookup table
		m_stateLUT[s.toString()] = s;
		
		// Create the transition distribution for this state
		m_transitions[s] = new DiscreteDistribution;
	}

	//
	// Get a state by name
	//
	State getState(char[] name)
	{
		State* pState = (name in m_stateLUT);
		if(pState == null)
		{
			throw new Exception("State " ~ name ~ " does not exist");
		}
		return *pState;
	}

	//
	// Get the transition probablity distribution for the given state
	//
	DiscreteDistribution getTransitionWeights(State s)
	{
		// Ensure the state exists
		DiscreteDistribution* pDist = (s in m_transitions);
		if(pDist == null)
		{
			throw new Exception("State " ~ s.toString ~ " does not exist in the markov model");
		}
		return *pDist;
	}

	//
	//
	//
	bool hasTransition(State k, State l)
	{
		IDistribution dist = getTransitionWeights(k);
		return dist.contains(l);
	}

	//
	//
	//
	void createTransition(State s1, State s2, real weight = 0.0f)
	{
		DiscreteDistribution dist = getTransitionWeights(s1);
		try
		{
			dist.addSymbol(s2);
			dist.setWeight(s2,weight);
		}
		catch(Exception e)
		{
			Stdout("Expection in create transition: ")(e);
		}
	}

	//
	//
	//
	void normalizeTransitions(State s)
	{
		getTransitionWeights(s).normalize();
	}

	//
	//
	//
	void printAll()
	{
		// Print the state transitions
		Stdout("Transitions:").newline;
		foreach(s,d; m_transitions)
		{
			Stdout.formatln("{0} -> {1:F6}", s, d);
		}

		Stdout.newline;
		Stdout("Emission Distributions:").newline;
		foreach(s,d; m_transitions)
		{
			Stdout(s)(" = ")(s.getEmissionDistribution()).newline;
		}
	}

	//
	// Write out the model in DOT format
	// 
	void writeDot(char[] filename)
	{
		TextFileOutput outFile = new TextFileOutput(filename);
		outFile("digraph G {\n");
		
		State[] states = getStateList();

		foreach(s; states)
		{
			if(s == getTerminalState())
			{
				continue;
			}
			outFile.formatln("\"{0}\" [ label=\"{1}\" ];", s, s);
		}

		// Print the transitions for each state
		foreach(s; states)
		{
			if(s == getTerminalState())
			{
				continue;
			}

			DiscreteDistribution transitions = getTransitionWeights(s);

			foreach(t; states)
			{
				if(t == getTerminalState())
				{
					continue;
				}

				if(hasTransition(s,t))
				{
					outFile.formatln("\"{0}\" -> \"{1}\" [ label=\"{2:F3}\" ];", s, t, transitions.getWeight(t));
				}
			}
		}
		outFile("}\n");
		outFile.flush();
		outFile.close();
	}

	//
	//
	//
	Symbol[] getEmissionList() { return m_symbolAlphabet.getSymbols(); }
	
	//
	//
	//
	State[] getStateList() { return m_transitions.keys; }
	
	//
	//
	//
	State getTerminalState() { return m_terminal; }
	

	//
	// Members
	//
	State[char[]] m_stateLUT;
	DiscreteDistribution[State] m_transitions;
	State m_terminal;
	IAlphabet m_symbolAlphabet;
}
